@@ -59,6 +59,27 @@ query WorkerWaitingJobs($worker: String){
 }
 '''
 
+paginate_stale_jobs = '''
+  query WorkerWaitingJobsPaginate(
+    $worker: String!
+    $cursor: Int = 0
+    $limit: Int = 10
+  ){
+    jobsPaginate(
+      cursor: $cursor
+      filter: {
+        workerCode: $worker
+        jobStatus: WAITING
+      }
+      orderBy: {asc: ID}
+      limit: $limit
+    ) {
+      pageInfo { endCursor hasNextPage }
+      jobs: edge{ id }
+    }
+  }
+'''
+
 get_job ='''
 query GetJob(
   $id: ID!
@@ -103,7 +124,7 @@ query QueueByName(
 }
 '''
 
-all_queues = '''
+all_queues_w_listeners = '''
 {
   queues{
     id
@@ -111,18 +132,29 @@ all_queues = '''
     alias
     description
     context
-    enabled
     type
     schedule
     query
     lastRunAt
     lockRequired
+    jobs(orderBy: { desc:ID } limit: 1){
+      id
+      updatedAt
+    }
     locks(orderBy:{ desc:ID } limit: 1){
       id
       active
       startAt
       endAt
     }
+    listeners: queueWorkers {
+      id
+      enabled
+      worker{
+        id
+        code
+      }
+    }
     updatedAt
     insertedAt
   }
@@ -223,3 +255,20 @@ get_group_users = '''
     }
   }
 '''
+
+get_worker_queues = '''
+  query WorkerQueues($workerCode: String!){
+    queueWorkers(filter: {workerCode: $workerCode}){
+      id
+      worker{
+        id
+        code
+      }
+      queue{
+        id
+        name
+      }
+      enabled
+    }
+  }
+'''
