@@ -1,4 +1,5 @@
 import pydash as __
+from pprint import pformat
 import time
 import json
 from json.decoder import JSONDecodeError
@@ -35,6 +36,10 @@ class PollingMode(Enum):
   QUERY = 'QUERY'
   SUBSCRIPTION = 'SUBSCRIPTION'
 
+class JobConfigMode(Enum):
+  KEEP = 'KEEP' # do not touch anything at startup
+  SYNC = 'SYNC' # sync Queues, job configuration, remote logging configuration, etc
+
 DEFAULT_EVENT_CONFIG = {
   "timeout": 5.0, # seconds to wait, None to run ASAP
   "min_batch": 10,
@@ -53,12 +58,14 @@ class ValiotWorker():
     self.queues = {}
     self.eventQueues = {}
     self.currentQueues = []
+    self.previousQueues = []
     self.runningJobs = {}
     self.lockedJobs = []
     self.gql = GraphQLClient()
     self.worker = None
     self.worker_code = None
     self.pollingMode = PollingMode.QUERY
+    self.jobConfigMode = JobConfigMode.KEEP
     self.context = None
     self.unsubJobCreated = None
     # logging attributes
@@ -79,6 +86,9 @@ class ValiotWorker():
   def setPollingMode(self, mode):
     self.pollingMode = mode
   
+  def setJobConfigMode(self, mode):
+    self.jobConfigMode = mode
+  
   def setLoggingStyle(self, style):
     self.loggingStyle = style
 
@@ -88,7 +98,7 @@ class ValiotWorker():
     alias, 
     description='', 
     schedule='', 
-    enabled=None, 
+    enabled=None, # ! PER-WORKER BASIS
     queueType=QueueType.ON_DEMAND,
     query='',
     lockRequired=False,
@@ -164,8 +174,99 @@ class ValiotWorker():
       "last_run_at": datetime.now()
     }
     fn_name = __.get(job, 'queue.name', 'NOT_FOUND')
-    job_dict = __.get(self.queues, fn_name, NOT_FOUND)
+    job_dict = __.get(self.queues, fn_name, None)
     return fn_name, job_dict
+  
+  def groupQueuesByListenStatus(self, queues):
+    not_listened_queues = [
+        queue['name']
+        for queue in queues
+        if not __.get(queue, 'listeners', [])
+    ]
+    worker_listened_queues = [
+        queue['name']
+        for queue in queues
+        if __.find(
+          __.get(queue, 'listeners', []),
+          lambda listener: (
+            self.worker_code == __.get(listener, 'worker.code') and
+            __.get(listener, 'enabled')
+          )
+        )
+    ]
+    worker_disabled_queues = [
+      queue['name']
+      for queue in queues
+      if __.find(
+        __.get(queue, 'listeners', []),
+        lambda listener: (
+          self.worker_code == __.get(listener, 'worker.code') and
+          not __.get(listener, 'enabled')
+        )
+      )
+    ]
+    return (
+      sorted(worker_listened_queues),
+      sorted(worker_disabled_queues),
+      sorted(not_listened_queues),
+    )
+  
+  def reportListenedQueuesExhaustive(self):
+    (
+      listened,
+      disabled,
+      not_listened
+    ) = self.groupQueuesByListenStatus(self.currentQueues)
+    self.log(LogLevel.SUCCESS, 'LISTENED Queues:')
+    for queue_name in listened:
+      self.log(LogLevel.SUCCESS, f"\t{queue_name}")
+    if not listened:
+      self.log(LogLevel.SUCCESS, f'{self.worker_code} not listening to any Queue')
+    self.log(LogLevel.WARNING, 'DISABLED Queues:')
+    for queue_name in disabled:
+      self.log(LogLevel.WARNING, f"\t{queue_name}")
+    if not disabled:
+      self.log(LogLevel.WARNING, f'\tNo queues disabled')
+    self.log(LogLevel.ERROR, 'NOT LISTENED Queues:')
+    for queue_name in not_listened:
+      self.log(LogLevel.ERROR, f"\t{queue_name}")
+    if not not_listened:
+      self.log(LogLevel.INFO, f'\tAll queues listened')
+  
+  def reportListenedQueuesDiff(self):
+    (
+      listened,
+      disabled,
+      not_listened
+    ) = self.groupQueuesByListenStatus(self.currentQueues)
+    (
+      prev_listened,
+      prev_disabled,
+      prev_not_listened
+    ) = self.groupQueuesByListenStatus(self.previousQueues)
+    listened_diff = set(listened) - set(prev_listened)
+    disabled_diff = set(disabled) - set(prev_disabled)
+    not_listened_diff = set(not_listened) - set(prev_not_listened)
+    if listened_diff:
+      self.log(LogLevel.SUCCESS, f'NEW LISTENED QUEUES:')
+      for queue_name in listened_diff:
+        self.log(LogLevel.SUCCESS, f'\t{queue_name}')
+    if disabled_diff:
+      self.log(LogLevel.WARNING, f'NEW DISABLED QUEUES:')
+      for queue_name in disabled_diff:
+        self.log(LogLevel.WARNING, f'\t{queue_name}')
+    if not_listened_diff:
+      self.log(LogLevel.ERROR, f'NEW NOT LISTENED QUEUES:')
+      for queue_name in not_listened_diff:
+        self.log(LogLevel.ERROR, f'\t{queue_name}')
+
+  def reportQueueChanges(self):
+    if not self.previousQueues:
+      # first iteration, report exhaustive lists (Listening, skipping, disabled)
+      self.reportListenedQueuesExhaustive()
+    else:
+      # Check for diffs, and only report that
+      self.reportListenedQueuesDiff()
 
   def abortJob(self, job):
     variables = {
@@ -206,6 +307,11 @@ class ValiotWorker():
       lock = self.lockQueue(job["queue"]["name"])
     # Obtiene job
     fn_name, job_dict = self.getJobData(job)
+    # ! Validates existence of requested job for this worker:
+    if not job_dict:
+      self.log(LogLevel.ERROR, f"{fn_name} job not found (it might exist in another branch / parallel universe), skipping...")
+      # TODO: Here we can update the job to aborted or something
+      return
     self.log(LogLevel.INFO, f"running {fn_name}")
     job_fn = job_dict['function']
     # Obtiene inputs
@@ -308,6 +414,25 @@ class ValiotWorker():
     if not __.get(current_queue, 'lockRequired'):
       return False # Queue does not require locking
     return __.get(current_queue, 'locks.0.active') # locked if last lock is active
+  
+  def workerQueueEnabled(self, queue):
+    # queue not enabled if: no listener for this queue OR current worker not listening OR listening disabled for this worker
+    currentQueue = __.find(
+      self.currentQueues,
+      lambda cq: queue.get('name') and queue['name'] == cq['name']
+    )
+    listeners = __.get(currentQueue, 'listeners', [])
+    if not listeners:
+      return False
+    listener = __.find(
+      listeners,
+      lambda lnr: self.worker_code == __.get(lnr, 'worker.code')
+    )
+    if not listener:
+      return False
+    if __.get(listener, 'enabled'):
+      return True
+    return False
 
   def processJobs(self):
     jobs, errors = self.gql.query(queries.all_jobs, variables={'worker': self.worker['code']})
@@ -321,7 +446,7 @@ class ValiotWorker():
         if self.queue_locked(job['queue']['name']):
           self.log(LogLevel.ERROR, f"{job['queue']['name']} is LOCKED, job aborted")
           self.abortJob(job)
-        elif job['queue']['enabled']:
+        elif self.workerQueueEnabled(job['queue']):
           self.runJob(job)
         else:
           self.log(LogLevel.WARNING, f"{job['queue']['name']} is DISABLED, skipping job")
@@ -329,22 +454,45 @@ class ValiotWorker():
       self.log(LogLevel.INFO, "Listening to events...")
 
   def abortStaleJobs(self):
-    jobs, errors = self.gql.query(queries.waiting_jobs, variables={'worker': self.worker['code']})
-    if (errors):
-      self.log(LogLevel.ERROR, f'errors: {errors}')
-    elif (len(jobs) > 0):
-      self.log(LogLevel.ERROR, f'Deleting {len(jobs)} stale jobs:')
-      for job in jobs:
-        result, errors = self.gql.mutate(mutations.delete_job, {'id': job['id']})
-        if (not errors):
-          self.log(LogLevel.ERROR, f'deleted stale job {job["id"]}:')
-        else:
-          self.log(LogLevel.ERROR, f'errors deleting stale job: {errors}')
+    # * pagination configuration
+    workerRegex = f'^{self.worker["code"]}$' # worker exact match
+    limit = 100 # how many jobs to delete in one batch
+    cursor = 0 # starting point for each page
+    hasNextPage = True
+    abortedCount = 0
+    pages = 0
+    while hasNextPage:
+      # * Get staled jobs page
+      data, errors = self.gql.query(
+        queries.paginate_stale_jobs,
+        variables={'worker': workerRegex, 'cursor': cursor, 'limit': limit}
+      )
+      jobs = __.get(data, 'jobs', [])
+      cursor = __.get(data, 'pageInfo.endCursor')
+      hasNextPage = __.get(data, 'pageInfo.hasNextPage')
+      # ! delete 'em jobs
+      # * get batch mutation
+      BATCH_STALE_DELETE_MUT = mutations.build_batch_stale_mutation(jobs)
+      if not BATCH_STALE_DELETE_MUT:
+        continue
+      # * run mutation
+      self.log(LogLevel.WARNING, f'Deleting {len(jobs)} stale jobs...')
+      data, errors = self.gql.mutate(BATCH_STALE_DELETE_MUT)
+      if errors:
+        self.log(LogLevel.ERROR, 'Errors deleting stale jobs:')
+        self.log(LogLevel.ERROR, errors)
+      self.log(LogLevel.ERROR, f'Deleted {len(jobs)} stale jobs')
+      abortedCount += len(jobs)
+      pages += 1
+    if abortedCount and pages >= 2:
+      self.log(LogLevel.ERROR, f'Deleted a total of {abortedCount} stale jobs')
+    elif not abortedCount:
+      self.log(LogLevel.INFO, 'No stale jobs to delete')
 
   def updateAvailableQueues(self):
     self.log(LogLevel.SUCCESS, 'Actualizando Jobs disponibles')
+    # ! Update each queue / config parameter
     for name, queue in self.queues.items():
-      enabled = __.get(queue, 'enabled', None)
       description = __.get(queue, 'description', None)
       schedule = __.get(queue, 'schedule', None)
       query = __.get(queue, 'query', None)
@@ -356,14 +504,13 @@ class ValiotWorker():
         'lockRequired': lockRequired
       }
       # ! Add optional variables
-      if enabled is not None:
-        variables['enabled'] = enabled
       if description:
         variables['description'] = description
       if schedule:
         variables['schedule'] = schedule
       if query:
         variables['query'] = query
+      # ! Update QUEUE-related config
       data, errors = self.gql.mutate(
         mutation=mutations.upsert_queue,
         variables=variables
@@ -373,7 +520,42 @@ class ValiotWorker():
         self.log(LogLevel.ERROR, errors)
       else:
         self.log(LogLevel.SUCCESS, f'Queue {name} actualizado correctamente')
-
+  
+  def updateWorkerQueues(self):
+    # ! Update WORKER-related config
+    # * first, get current configuration (may cause bugs if we create more than one config register, see Many-Many)
+    workerQueues, errors = self.gql.query(
+      query=queries.get_worker_queues,
+      variables={'workerCode': self.worker_code}
+    )
+    if errors:
+      self.log(LogLevel.ERROR, errors)
+    # ! Manual Upsert (Not enabled for many-many relationships)
+    for name, queue in self.queues.items():
+      # ! check whether we should create or update the item:
+      workerQueue = __.find(
+        workerQueues, 
+        lambda wk: name == __.get(wk, 'queue.name', None)
+      )
+      queueEnabled = __.get(queue, 'enabled', False)
+      if not workerQueue:
+        _, errors = self.gql.mutate(mutations.create_queue_worker, { 'worker': self.worker_code, 'queue': name, 'enabled': queueEnabled })
+        if errors:
+          log_lvl = LogLevel.ERROR
+          log_msg = f'Error Creating config for [ {self.worker_code} <=> {name} ]: {errors}'
+        else:
+          log_lvl = LogLevel.SUCCESS if queueEnabled else LogLevel.WARNING
+          log_msg = f'Created config for [ {self.worker_code} <=> {name} ] = {"enabled" if queueEnabled else "disabled"}'
+      else:
+        _, errors = self.gql.mutate(mutations.update_queue_worker, { 'id': workerQueue['id'], 'enabled': queueEnabled })
+        if errors:
+          log_lvl = LogLevel.ERROR
+          log_msg = f'Error Updating config for [ {self.worker_code} <=> {name} ]: {errors}'
+        else:
+          log_lvl = LogLevel.SUCCESS if queueEnabled else LogLevel.WARNING
+          log_msg = f'Updated config for [ {self.worker_code} <=> {name} ] = {"enabled" if queueEnabled else "disabled"}'
+      self.log(log_lvl, log_msg)
+  
   def clearFinishedJobs(self):
     finished = []
     for job_id, running_job in self.runningJobs.items():
@@ -406,18 +588,25 @@ class ValiotWorker():
         self.runEvent(queue['name'])
   
   def runFrequencyQueues(self):
+    """
+    This function checks if it has elapsed enough time since the last attempt to run this queue.
+    It takes into account the last time a job of this queue was created,
+    rather than when was the last time the job RAN, this enforces only 1 attempt per interval.
+    """
     from pytz import timezone
     utc = timezone("UTC")
     queuesByType = __.group_by(self.currentQueues, 'type')
     freqQueues = __.get(queuesByType, 'FREQUENCY', default=[])
     # self.log(LogLevel.WARNING, f'checking {len(freqQueues)} frequency queues')
     for queue in freqQueues:
-      # if it has never been ran, set the lastRunAt as a very old date (forces next run)
-      lastRunAt = __.get(queue, 'lastRunAt', None)
-      if not lastRunAt:
-        lastRunDate = datetime(2000, 1, 1)
-      else:
-        lastRunDate = datetime.strptime(lastRunAt, "%Y-%m-%dT%H:%M:%SZ")
+      lastRunAt = __.get(
+        queue,
+        # ! get from job.updatedAt, rather than queue.lastRunAt itself (see description of this method)
+        'jobs.0.updatedAt',
+        # ! Default: if it has never been ran, set the lastRunAt as a very old date (forces next run)
+        "2000-01-01T00:00:00Z"
+      )
+      lastRunDate = datetime.strptime(lastRunAt, "%Y-%m-%dT%H:%M:%SZ")
       lastRunDate = utc.localize(lastRunDate)
       now = getUtcDate(datetime.now())
       cron_iter = get_croniter(queue['schedule'], now)
@@ -427,7 +616,7 @@ class ValiotWorker():
       freq = (nextRunAt[1] - nextRunAt[0]).total_seconds() / 60.0 # in minutes
       if (elapsedFromLastRun > freq):
         # ! Run the frequency job!
-        if (queue["enabled"]):
+        if self.workerQueueEnabled(queue):
           self.gql.mutate(
             mutation=mutations.create_job,
             variables={
@@ -448,7 +637,7 @@ class ValiotWorker():
       if self.queue_locked(job['queue']['name']):
         self.log(LogLevel.ERROR, f"{job['queue']['name']} is LOCKED, job aborted")
         self.abortJob(job)
-      elif job['queue']['enabled']:
+      elif self.workerQueueEnabled(job['queue']):
         self.runJob(job)
       else:
         self.log(LogLevel.WARNING, f"{job['queue']['name']} is DISABLED, skipping job")
@@ -463,14 +652,36 @@ class ValiotWorker():
       self.log(LogLevel.ERROR, 'ERROR SUBSCRIBING TO JOB CREATIONS')
   
   def runEvent(self, queue_name):
+    '''
+    This wrapper ensures that whenever runEvent is called, the event is cleared
+    '''
+    cleanup = self.actuallyRunEvent(queue_name)
+    # Move mailbox data and clear it
+    if cleanup:
+      self.eventQueues[queue_name]["mailbox"] = []
+      self.eventQueues[queue_name]["start_time"] = None # Clear timeout
+
+  def actuallyRunEvent(self, queue_name):
+    """Runs job of type ON_EVENT
+
+    Parameters
+    ----------
+    queue_name : string
+      Name of the queue to run
+    Returns
+    -------
+    boolean
+      True if a cleanup should be performed (event cleanup)
+      False if cleanup should be skipped (to run event at a later point)
+    """
     queuesByName = __.group_by(self.currentQueues, 'name')
     currentQueue = __.get(queuesByName, f'{queue_name}.0')
     if self.queue_locked(currentQueue['name']):
       self.log(LogLevel.WARNING, f"{currentQueue['name']} is LOCKED, event postponed")
-      return
-    elif not currentQueue['enabled']:
+      return False
+    elif not self.workerQueueEnabled(currentQueue):
       self.log(LogLevel.WARNING, f'Event triggered for {queue_name} but Queue not enabled, skipping...')
-      return
+      return True
     data = self.eventQueues[queue_name]["mailbox"]
     # ! try to run job
     # obtiene lock (if needed):
@@ -488,7 +699,7 @@ class ValiotWorker():
     )
     if errors:
       self.log(LogLevel.ERROR, f'Error starting event:\n{errors}')
-      return
+      return True
     job = __.get(response, 'job.result')
     def update_this_job(variables):
       # * funcion de utilidad para actualizar job sin necesidad de conocer su ID
@@ -548,10 +759,7 @@ class ValiotWorker():
         'job': {**job, 'queue': {**currentQueue}},
         'lock': lock
       })
-    # Move mailbox data and clear it
-    self.eventQueues[queue_name]["mailbox"] = []
-    self.eventQueues[queue_name]["start_time"] = None # Clear timeout
-
+    return True
   def registerEventJobs(self):
     queuesByType = __.group_by(self.queues.values(), 'type')
     eventQueues = __.get(queuesByType, 'EVENT', default=[])
@@ -566,6 +774,8 @@ class ValiotWorker():
         data : any
             Data received from the event
         """
+        if not self.workerQueueEnabled(queue):
+          return  # Skipping event
         if len(self.eventQueues[queue["name"]]["mailbox"]) == 0:
           self.log(LogLevel.WARNING, f'EVENT { queue["name"] } TRIGGERED, BUILDING BATCH')
           self.eventQueues[queue["name"]]["start_time"] = time.time()
@@ -665,11 +875,12 @@ class ValiotWorker():
 
   def eventLoop(self, interval=1.0):
     try:
-      currentQueues, errors = self.gql.query(queries.all_queues)
+      currentQueues, errors = self.gql.query(queries.all_queues_w_listeners)
       if (errors):
         self.log(LogLevel.ERROR, f'errors: {errors}')
         return
       self.currentQueues = currentQueues
+      self.reportQueueChanges()
       self.clearFinishedJobs()
       self.unlockFinishedJobs()
       self.runFrequencyQueues()
@@ -682,6 +893,7 @@ class ValiotWorker():
     except Exception as e:
       self.log(LogLevel.ERROR, f"Error in worker's event loop!")
       self.log(LogLevel.ERROR, f'e: {e}')
+    self.previousQueues = self.currentQueues
     time.sleep(interval)
   
   def getWorker(self):
@@ -699,7 +911,11 @@ class ValiotWorker():
     self.getWorker()
     self.startLoggingLoop()
     self.abortStaleJobs()
-    self.updateAvailableQueues()
+    if (self.jobConfigMode == JobConfigMode.SYNC):
+      self.updateAvailableQueues()
+      self.updateWorkerQueues()
+    else:
+      self.log(LogLevel.INFO, 'No config sync performed. If required, set jobConfigMode to SYNC')
     if (self.pollingMode == PollingMode.SUBSCRIPTION):
       self.registerJobCreatedSubscription()
     self.registerEventJobs()
